/* Copyright (C) 2010 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the src/utils sub-directory);
   if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file or uses macros or
inline functions (of any length) should by reason only of that
instantiation or use be subject to the restrictions of use in the GNU
Lesser General Public License.  With that in mind, the words "and
macros, inline functions and instantiations of templates (of any
length)" shall be treated as substituted for the words "and small
macros and small inline functions (ten lines or less in length)" in
the fourth paragraph of section 5 of that licence.  This does not
affect any other reason why object code may be subject to the
restrictions in that licence (nor for the avoidance of doubt does it
affect the application of section 2 of that licence to modifications
of the source code in this file).

*/

#ifndef FUTURE_H
#define FUTURE_H

#include <memory>
#include <exception>

#include <pthread.h>

#include <c++-gtk-utils/thread.h>
#include <c++-gtk-utils/mutex.h>
#include <c++-gtk-utils/callback.h>
#include <c++-gtk-utils/intrusive_ptr.h>
#include <c++-gtk-utils/emitter.h>


namespace Thread {

struct FutureThreadError: public std::exception {
  virtual const char* what() const throw() {return "FutureThreadError\n";}
};


/**
 * @class Thread::Future future.h c++-gtk-utils/future.h
 * @brief A class representing a pthread thread which will
 * provide a value.
 * @sa Thread::Thread Thread::JoinableHandle
 *
 * The Thread::Future class will launch a worker thread, run the
 * function it represents in that thread until it returns, and store
 * the return value so that it can be waited on and/or extracted by
 * another thread.  A new Thread::Future object representing the
 * function to be called is created by calling make().  The worker
 * thread is then started by calling run(), and the value extracted or
 * waited for by calling get().  The run() method can only be called
 * once, but any number of threads can wait for and/or extract the
 * return value by calling the get() method.  The class also provides
 * a SafeEmitter @ref DoneEmitterAnchor "done_emitter" public object
 * which emits when the worker thread has finished (see the
 * documentation on the use of @ref DoneEmitterAnchor "done_emitter"
 * below).
 *
 * The template parameter type is the type of the return value of the
 * function called by the Thread::Future object.  The return value can
 * be any type, including any arbitrarily large tuple or other struct
 * or standard C++ container.
 *
 * A Thread::Future object cannot represent a function with a void
 * return type - a compilation error will result if that is attempted.
 * If no return value is wanted, then the Thread::Thread class can be
 * used directly.  (However, if in a particular usage this class is
 * thought to be more convenient, the function to be represented by it
 * can be wrapped by another function which provides a dummy return
 * value, such as a dummy int.  One possible case for this is where
 * more than one thread wants to wait for the worker thread to
 * terminate, as pthread_join() and so Thread::Thread::join() only
 * give defined behaviour when called by one thread.)
 * 
 * The make() method takes a plain function, static member function or
 * non-static member function, and the function can take up to two
 * arguments.  It works in the same way as does the Callback::make()
 * factory function, except that Thread::Future<>::make() returns the
 * Thread::Future object by IntrusivePtr<Thread::Future<> >.
 *
 * This is a usage example:
 *
 * @code 
 *   class Numbers {
 *   public:
 *     std::vector<long> get_primes(int n); // calculates the first n primes
 *                                          // and puts them in a vector
 *    ...
 *   };
 *
 *   Numbers obj;
 *
 *   // get the first 1,000 primes
 *   IntrusivePtr<Thread::Future<std::vector<long> > > future = 
 *     Thread::Future<std::vector<long> >::make(obj, &Numbers::get_primes, 1000));
 *   future->run();
 *   ... [ do something else ] ...
 *   std::vector<long> result(future->get());
 *   for (std::vector<long>::const_iterator i = result.begin(),
 *        i != result.end(),
 *        ++i)
 *     std::cout << *i << std::endl;
 * @endcode
 *
 * If get_primes() were a static member function or plain function,
 * the syntax would be:
 *
 * @code 
 *   IntrusivePtr<Thread::Future<std::vector<long> > > future = 
 *     Thread::Future<std::vector<long> >::make(&Numbers::get_primes, 100));
 * @endcode
 *
 */

template <class Val>
class Future: public IntrusiveLockCounter {

  std::auto_ptr<Thread::Thread> thread_a;
  std::auto_ptr<Callback::Callback> cb_a;

  mutable Mutex mutex;
  Cond cond;
  Val val;
  bool done;
  bool running;
  bool error;
  bool emitter_error;

#ifndef DOXYGEN_PARSING
  /* we don't want to use std::pair as an alternative to arg_pair, as
     std::pair initialises its elements in its constructor by const
     reference and so may inhibit copy elision optimisation, and in
     addition it means we cannot pass a reference to the target
     object, which we do want to be held by (non-const) reference.
  */
  template<class R1, class R2>
  struct arg_pair {
    R1 r1;
    R2 r2;
    arg_pair(R1 r1_, R2 r2_):
      r1(r1_), r2(r2_) {}
  };

  template<class R1, class R2, class R3>
  struct arg_triple {
    R1 r1;
    R2 r2;
    R3 r3;
    arg_triple(R1 r1_, R2 r2_, R3 r3_):
      r1(r1_), r2(r2_), r3(r3_) {}
  };

  template<class R1, class R2, class R3, class R4>
  struct arg_quadruple {
    R1 r1;
    R2 r2;
    R3 r3;
    R4 r4;
    arg_quadruple(R1 r1_, R2 r2_, R3 r3_, R4 r4_):
      r1(r1_), r2(r2_), r3(r3_), r4(r4_) {}
  };
#endif

 /* All the run_wrapper() methods are static.  This is because gcc-4.2
  * has problems with instantiating a non-static nested template
  * member function from a static nested template member function
  * (Future<>::make())
  */
  template <class Ret, class T>
  static void run_wrapper(arg_pair<T&, Ret (T::*)()> args,
			  Future<Val>* instance);

  template <class Ret, class Arg1, class T>
  static void run_wrapper(arg_triple<T&, Ret (T::*)(Arg1), Arg1> args,
			  Future<Val>* instance);

  template <class Ret, class Arg1, class Arg2, class T>
  static void run_wrapper(arg_quadruple<T&, Ret (T::*)(Arg1, Arg2), Arg1, Arg2> args,
			  Future<Val>* instance);

  template <class Ret, class T>
  static void run_wrapper_const(arg_pair<const T&, Ret (T::*)() const> args,
				Future<Val>* instance);

  template <class Ret, class Arg1, class T>
  static void run_wrapper_const(arg_triple<const T&, Ret (T::*)(Arg1) const, Arg1> args ,
				Future<Val>* instance);

  template <class Ret, class Arg1, class Arg2, class T>
  static void run_wrapper_const(arg_quadruple<const T&, Ret (T::*)(Arg1, Arg2) const, Arg1, Arg2> args,
				Future<Val>* instance);

  template <class Ret>
  static void run_wrapper_static(Ret (*func)(),
				 Future<Val>* instance);

  template <class Ret, class Arg1>
  static void run_wrapper_static(arg_pair<Ret (*)(Arg1), Arg1> args,
				 Future<Val>* instance);

  template <class Ret, class Arg1, class Arg2>
  static void run_wrapper_static(arg_triple<Ret (*)(Arg1, Arg2), Arg1, Arg2> args,
				 Future<Val>* instance);

  // private constructor - this class can only be created with Thread::Future::make()
  Future(): val(), done(false), running(false), error(false), emitter_error(false) {}
  // and it cannot be copied except by smart pointer
  Future(const Future&);
  Future& operator=(const Future&);

  void cancel_cleanup();

public:

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the default constructor of the return value type throws.
 *
 * Since 1.0.2
 */
  template <class Ret, class T>
  static IntrusivePtr<Future<Val> > make(T& t,
					 Ret (T::*func)());

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the copy constructor or assignment operator of the bound
 * argument throws, or the default constructor of the return value
 * type throws.
 *
 * Since 1.0.2
 */
  template <class Ret, class Arg1, class T>
  static IntrusivePtr<Future<Val> > make(T& t,
					 Ret (T::*func)(Arg1),
					 Arg1 arg1);

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the copy constructor or assignment operator of a bound
 * argument throws, or the default constructor of the return value
 * type throws..
 *
 * Since 1.0.2
 */
  template <class Ret, class Arg1, class Arg2, class T>
  static IntrusivePtr<Future<Val> > make(T& t,
					 Ret (T::*func)(Arg1, Arg2),
					 Arg1 arg1,
					 Arg2 arg2);

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the default constructor of the return value type throws.
 *
 * Since 1.0.2
 */
  template <class Ret, class T>
  static IntrusivePtr<Future<Val> > make(const T& t,
					 Ret (T::*func)() const);

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the copy constructor or assignment operator of the bound
 * argument throws, or the default constructor of the return value
 * type throws..
 *
 * Since 1.0.2
 */
  template <class Ret, class Arg1, class T>
  static IntrusivePtr<Future<Val> > make(const T& t,
					 Ret (T::*func)(Arg1) const,
					 Arg1 arg1);

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the copy constructor or assignment operator of a bound
 * argument throws, or the default constructor of the return value
 * type throws.
 *
 * Since 1.0.2
 */
  template <class Ret, class Arg1, class Arg2, class T>
  static IntrusivePtr<Future<Val> > make(const T& t,
					 Ret (T::*func)(Arg1, Arg2) const,
					 Arg1 arg1,
					 Arg2 arg2);

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the default constructor of the return value type throws.
 *
 * Since 1.0.2
 */
  template <class Ret>
  static IntrusivePtr<Future<Val> > make(Ret (*func)());

/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the copy constructor or assignment operator of the bound
 * argument throws, or the default constructor of the return value
 * type throws..
 *
 * Since 1.0.2
 */
  template <class Ret, class Arg1>
  static IntrusivePtr<Future<Val> > make(Ret (*func)(Arg1),
					 Arg1 arg1);
/**
 * Constructs a new Thread::Future object (returned by
 * IntrusivePtr<Thread::Future<Val> >).  The type parameter Val
 * represents the return value of the function to be represented by
 * the new object.
 * @exception std::bad_alloc It might throw std::bad_alloc if memory
 * is exhausted and the system throws in that case.  It will also
 * throw if the copy constructor or assignment operator of a bound
 * argument throws, or the default constructor of the return value
 * type throws..
 *
 * Since 1.0.2
 */
  template <class Ret, class Arg1, class Arg2>
  static IntrusivePtr<Future<Val> > make(Ret (*func)(Arg1, Arg2),
					 Arg1 arg1,
					 Arg2 arg2);
/**
 * Runs the function represented by this Thread::Future object in a
 * new worker thread.  That function will only be run once.  If this
 * is the first time this method has been called, it will start the
 * worker thread and return true, and if it has previously been
 * called, this method will do nothing and return false.  This method
 * is thread safe and may be called by a different thread from the one
 * which called make().
 * @return true if this is the first time this method has been called,
 * or false if this method has previously been called.
 * @exception Thread::FutureThreadError This method might throw
 * Thread::FutureThreadError if it has not previously been called and
 * the thread did not start properly.
 * @exception std::bad_alloc This method might throw std::bad_alloc if
 * it has not previously been called, and memory is exhausted and the
 * system throws in that case.
 * @note 1. Any Thread::Exit exception, or any uncaught exception
 * derived from std::exception, which is thrown from the worker thread
 * will be caught and consumed and the error flag will be set.  The
 * worker thread will safely terminate and unwind the stack in so
 * doing.
 * @note 2. As this wrapper class can provide error reporting in a way
 * that Thread::Thread of itself cannot, it would be desirable to
 * consume any other uncaught exceptions.  However, this cannot be
 * done: annoyingly, NPTL's forced stack unwinding does not allow this
 * if thread cancellation is to be made available.  Neither POSIX nor
 * the current C++98/03 standard specifies what is to happen if an
 * uncaught exception propagates out of a thread.  In practice, most
 * implementations will call std::terminate() and so terminate the
 * entire program.  Accordingly, a user must make sure that no
 * exceptions, other than Thread::Exit or those derived from
 * std::exception or any cancellation pseudo-exception, can propagate
 * from the function which this Thread::Future object represents.
 * @note 3. If the worker thread is cancelled by a call to cancel()
 * while in the middle of executing the function which this
 * Thread::Future object represents, the error flag will be set.
 *
 * Since 1.0.2
 */
  bool run();

/**
 * Gets the value obtained from the function which is represented by
 * this object.  If the worker thread launched by the call to run()
 * has not completed, then this method will block until it has
 * completed.  If run() has not been called, then run() will be called
 * (and this method will block until the launched worker thread
 * completes).  If the function which is represented by this object
 * throws Thread::Exit or an uncaught exception derived from
 * std::exception, then the exception will have been consumed by this
 * Thread::Future object and the error flag will have been set.  The
 * error flag will also have been set if the worker thread is
 * cancelled or the thread wrapper in this Thread::Future object threw
 * std::bad_alloc.  This method is thread safe and may be called by
 * any thread (and by more than one thread).  It is also strongly
 * exception safe: no data will be lost if extracting the value fails.
 * @return The value obtained from the function which is represented
 * by this object
 * @exception Thread::FutureThreadError This method might throw
 * Thread::FutureThreadError if run() has not previously been called
 * and the thread did not start properly when this function called
 * run().
 * @exception std::bad_alloc This method might throw std::bad_alloc if
 * if run() has not previously been called, memory is exhausted and
 * the system throws in that case.
 * @note This method might also throw if the copy constructor or
 * assignment operator of the returned value type throws.
 *
 * Since 1.0.2
 */
  Val get();

/**
 * Cancels the worker thread in which the function represented by this
 * object runs, if that function has not yet finished.  If this method
 * is called and the worker thread is still running and is cancelled
 * in response to a call to this method, then the error flag will be
 * set so that a method calling get() can examine whether the result
 * is valid.  If run() has not yet been called or the worker thread
 * has already finished executing the function represented by this
 * object then this function does nothing and returns false.  This
 * method is thread safe and may be called by any thread.  It will not
 * throw.
 * @return true if run() has previously been called and the worker
 * thread has not yet finished executing the function represented by
 * this object, otherwise false (in which case this method does
 * nothing).
 * @note 1. Use this method with great care.  When cancelling a thread
 * not all thread implementations will unwind the stack, and so run
 * the destructors of local objects.  This is discussed further in the
 * documentation on Thread::Thread::cancel().
 * @note 2. This method might return true because the worker thread
 * has not yet finished, but the error flag might still not be set.
 * This is because the worker thread may not meet a cancellation point
 * before it ends naturally.  It is the error flag which indicates
 * definitively whether the worker thread terminated prematurely in
 * response to a call to this method.
 *
 * Since 1.0.2
 */
  bool cancel();

/**
 * @return true if the function represented by this Thread::Future
 * object has finished, either by returning normally, by cancellation
 * or by virtue of having thrown Thread::Exit or some exception
 * derived from std::exception.  This method is thread safe and may be
 * called by any thread.  It will not throw.
 *
 * Since 1.0.2
 */
  bool is_done() const;

/**
 * @return true if (a) a Thread::Exit exception has been thrown by the
 * function represented by this Thread::Future object (which will have
 * been consumed by this Thread::Future object), (b) an exception
 * derived from std::exception has been thrown by that function which
 * was not caught in that function (and which will also have been
 * consumed by this Thread::Future object), (c) the worker thread in
 * which that function runs was cancelled in mid-course with a call to
 * cancel() or (d) the thread wrapper implementing the worker thread
 * in this Thread::Future object threw std::bad_alloc (this is
 * different from the run() method throwing std::bad_alloc).  In these
 * cases the value obtained by get() will not be valid.  Otherwise
 * this method returns false.  This method is thread safe and may be
 * called by any thread.  It will not throw.
 *
 * Since 1.0.2
 */
  bool is_error() const;

/**
 * @return true if an uncaught exception arose in emitting @ref
 * DoneEmitterAnchor "done_emitter" when executing callbacks connected
 * to it.  Otherwise this method returns false.  This method is thread
 * safe and may be called by any thread.  It will not throw.
 *
 * Since 1.0.2
 */
  bool is_emitter_error() const;

/**
 * A SafeEmitter object which is emitted when the function represented
 * by this Thread::Future object finishes correctly (that is, the
 * function is not cancelled and does not throw any uncaught
 * exceptions).  By itself it does not do too much as it is emitted
 * (and connected callbacks execute in) the same worker thread
 * immediately after the Future function has completed.  However, any
 * thread can connect a Callback object to this SafeEmitter object and
 * a connected callback can, say, cause another Callback to be
 * executed in a thread's main loop using Callback::post().
 * @note 1. Cancellation is blocked while the SafeEmitter object emits
 * and any connected callback executes.
 * @note 2. A connected callback can however terminate the worker
 * thread by throwing Thread::Exit (in which case no subsequent
 * callbacks to be executed on that emission will execute either: the
 * worker thread will safely terminate and unwind the stack in so
 * doing).  In that event, the emitter_error flag will be set.
 * @note 3. All other uncaught exceptions which might be thrown by the
 * SafeEmitter object emitting, or by a connected callback function
 * executing, are consumed to retain the integrity of the
 * Thread::Future object.  In the event of such an exception being
 * thrown, the emitter_error flag will be set.  In summary, the
 * emitter_error flag will be set if (a) a callback function throws
 * Thread::Exit, (b) some other uncaught exception escapes from a
 * callback function or (c) SafeEmitter::emit() throws std::bad_alloc
 * or the assignment operator of a bound argument has thrown.  If the
 * user knows that the callback function does not throw Thread::Exit
 * and does not allow any other exception to escape, then the cause
 * must be a std::bad_alloc memory exception in SafeEmitter::emit() or
 * the assignment operator of a bound argument throwing.
 * @note 4. An emission is thread safe if the connected callback
 * functions are thread safe.
 * @anchor DoneEmitterAnchor
 *
 * Since 1.0.2
 */
  SafeEmitter done_emitter;
};

} // namespace Thread

#include <c++-gtk-utils/future.tcc>

#endif
