/* Copyright (C) 2010 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the src/utils sub-directory);
   if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file or uses macros or
inline functions (of any length) should by reason only of that
instantiation or use be subject to the restrictions of use in the GNU
Lesser General Public License.  With that in mind, the words "and
macros, inline functions and instantiations of templates (of any
length)" shall be treated as substituted for the words "and small
macros and small inline functions (ten lines or less in length)" in
the fourth paragraph of section 5 of that licence.  This does not
affect any other reason why object code may be subject to the
restrictions in that licence (nor for the avoidance of doubt does it
affect the application of section 2 of that licence to modifications
of the source code in this file).

*/

#ifndef FUTURE_TCC
#define FUTURE_TCC

namespace Thread {

extern "C" {
inline void cgu_future_cancel_cleanup_handler(void* arg) {
  Callback::Callback* cb = static_cast<Callback::Callback*>(arg);
  cb->dispatch();
  delete cb;
}
}

template <class Val>
void Future<Val>::cancel_cleanup() {
  {
    Mutex::Lock lock(mutex);
    error = true;
    done = true;
    cond.broadcast();  // any threads waiting on Future<Val>::get() will
                       // have their own reference, so we can unref() now
  }
  unref();
}

template <class Val>
template <class Ret, class T>
void Future<Val>::run_wrapper(arg_pair<T&, Ret (T::*)()> args, 
			      Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = ((args.r1).*(args.r2))();
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class Arg1, class T>
void Future<Val>::run_wrapper(arg_triple<T&, Ret (T::*)(Arg1), Arg1> args,
			      Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = ((args.r1).*(args.r2))(args.r3);
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class Arg1, class Arg2, class T>
  void Future<Val>::run_wrapper(arg_quadruple<T&, Ret (T::*)(Arg1, Arg2), Arg1, Arg2> args,
				Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = ((args.r1).*(args.r2))(args.r3, args.r4);
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class T>
void Future<Val>::run_wrapper_const(arg_pair<const T&, Ret (T::*)() const> args,
				    Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = ((args.r1).*(args.r2))();
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class Arg1, class T>
void Future<Val>::run_wrapper_const(arg_triple<const T&, Ret (T::*)(Arg1) const, Arg1> args,
				    Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = ((args.r1).*(args.r2))(args.r3);
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class Arg1, class Arg2, class T>
void Future<Val>::run_wrapper_const(arg_quadruple<const T&, Ret (T::*)(Arg1, Arg2) const, Arg1, Arg2> args,
				    Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = ((args.r1).*(args.r2))(args.r3, args.r4);
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret>
void Future<Val>::run_wrapper_static(Ret (*func)(),
				     Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = func();
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class Arg1>
void Future<Val>::run_wrapper_static(arg_pair<Ret (*)(Arg1), Arg1> args,
				     Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = args.r1(args.r2);
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

template <class Val>
template <class Ret, class Arg1, class Arg2>
void Future<Val>::run_wrapper_static(arg_triple<Ret (*)(Arg1, Arg2), Arg1, Arg2> args,
				     Future<Val>* instance) {
  Callback::Callback* cb;
  {
    CancelBlock b;
    try {
      cb = Callback::make(*instance, &Future<Val>::cancel_cleanup);
    }
    catch (...) {
      instance->cancel_cleanup();      
      return;
    }
  }

  pthread_cleanup_push(cgu_future_cancel_cleanup_handler, cb);

  try {
    // we don't want to hold the mutex when executing the callback (we may
    // be creating deadlocks from out of order locking) and we don't need
    // to when assigning to val, because before the get() method can proceed
    // this thread will have released the mutex below and the thread calling
    // get() will have since acquired it
    instance->val = args.r1(args.r2, args.r3);
  }
  catch (Exit&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  catch (std::exception&) {
    Mutex::Lock lock(instance->mutex);
    instance->error = true;
  }
  // we won't catch with the elypsis argument because NPTL forced stack
  // unwinding would require the exception to be rethrown.  For any exceptions
  // other than Thread::Exit and those derived from std::exception we will
  // have to rely on the user stopping them reaching this far.

  // don't allow any further thread cancellation, and particularly when
  // any callback connected to done_emitter is executed
  CancelBlock b;
  pthread_cleanup_pop(false);
  delete cb;

  instance->mutex.lock();
  instance->done = true;
  instance->cond.broadcast();
  if (!instance->error) {
    instance->mutex.unlock();
    try {
      instance->done_emitter();
    }
    catch (...) {
      Mutex::Lock lock(instance->mutex);
      instance->emitter_error = true;
    }
  }
  else instance->mutex.unlock();
  instance->unref();  // any threads waiting on Future<Val>::get() will
                      // have their own reference, so we can unref() now
}

  
template <class Val>
template <class Ret, class T>
IntrusivePtr<Future<Val> > Future<Val>::make(T& t,
					     Ret (T::*func)()) {
  
  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper<Ret, T>,
				      arg_pair<T&, Ret (T::*)()>(t, func),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class Arg1, class T>
IntrusivePtr<Future<Val> > Future<Val>::make(T& t,
					     Ret (T::*func)(Arg1),
					     Arg1 arg1) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper<Ret, Arg1, T>,
				      arg_triple<T&, Ret (T::*)(Arg1), Arg1>(t, func, arg1),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class Arg1, class Arg2, class T>
IntrusivePtr<Future<Val> > Future<Val>::make(T& t,
					     Ret (T::*func)(Arg1, Arg2),
					     Arg1 arg1,
					     Arg2 arg2) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper<Ret, Arg1, Arg2, T>,
				      arg_quadruple<T&, Ret (T::*)(Arg1, Arg2), Arg1, Arg2>(t, func, arg1, arg2),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class T>
IntrusivePtr<Future<Val> > Future<Val>::make(const T& t,
					     Ret (T::*func)() const) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper_const<Ret, T>,
				      arg_pair<const T&, Ret (T::*)() const>(t, func),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class Arg1, class T>
IntrusivePtr<Future<Val> > Future<Val>::make(const T& t,
					     Ret (T::*func)(Arg1) const,
					     Arg1 arg1) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper_const<Ret, Arg1, T>,
				      arg_triple<const T&, Ret (T::*)(Arg1) const, Arg1>(t, func, arg1),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class Arg1, class Arg2, class T>
IntrusivePtr<Future<Val> > Future<Val>::make(const T& t,
					     Ret (T::*func)(Arg1, Arg2) const,
					     Arg1 arg1,
					     Arg2 arg2) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper_const<Ret, Arg1, Arg2, T>,
				      arg_quadruple<const T&, Ret (T::*)(Arg1, Arg1) const, Arg1, Arg2>(t, func, arg1, arg2),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret>
IntrusivePtr<Future<Val> > Future<Val>::make(Ret (*func)()) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper_static<Ret>,
				      Ret (*func)(),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class Arg1>
IntrusivePtr<Future<Val> > Future<Val>::make(Ret (*func)(Arg1),
					     Arg1 arg1) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper_static<Ret, Arg1>,
				      arg_pair<Ret (*)(Arg1), Arg1>(func, arg1),
				      instance.get()));
  return instance;
}

template <class Val>
template <class Ret, class Arg1, class Arg2>
IntrusivePtr<Future<Val> > Future<Val>::make(Ret (*func)(Arg1, Arg2),
					     Arg1 arg1,
					     Arg2 arg2) {

  IntrusivePtr<Future<Val> > instance(new Future<Val>);
  instance->cb_a.reset(Callback::make(run_wrapper_static<Ret, Arg1, Arg2>,
				      arg_triple<Ret (*)(Arg1, Arg2), Arg1, Arg2>(func, arg1, arg2),
				      instance.get()));
  return instance;
}


template <class Val>
bool Future<Val>::run() {

  // take a reference so that this object will exist at least
  // until the worker thread has completed (we release in one
  // of the run_wrapper() methods, in cancel_cleanup() or on
  // throwing below)
  ref();

  mutex.lock();
  // we don't want to throw an exception if run() has already been
  // called - that may not be a programming error - the user may not
  // care whether it has been called, just that it should have been
  // called (Future<>::get() can be called any number of times by
  // any number of threads).  So provide a return value which can be
  // inspected where it is relevant to the particular usage case
  if (running) {
    mutex.unlock();
    unref();
    return false;
  }
  running = true;
  mutex.unlock();

  try {
    thread_a = Thread::Thread::start(cb_a.release(), false);
  }
  catch (...) {
    mutex.lock();
    error = true;
    mutex.unlock();
    unref();
    throw;
  }
  mutex.lock();
  if (!thread_a.get()) {
    error = true;
    mutex.unlock();
    unref();
    throw FutureThreadError();
  }
  mutex.unlock();
  return true;
}

template <class Val>
Val Future<Val>::get() {

  mutex.lock();
  if (!running) {   // make sure run() has been called
    mutex.unlock();
    run();          // there is no race - if another thread beats us
                    // to it after unlocking, run() will just return
    mutex.lock();
  }
  Mutex::Lock lock(mutex, locked);
  while (!done) cond.wait(mutex);
  return val;
}

template <class Val>
bool Future<Val>::cancel() {
  Mutex::Lock lock(mutex);
  if (running && !done && thread_a.get()) {
    thread_a->cancel();
    return true;
  }
  return false;     
}

template <class Val>
bool Future<Val>::is_done() const {
  Mutex::Lock lock(mutex);
  return done;
}

template <class Val>
bool Future<Val>::is_error() const {
  Mutex::Lock lock(mutex);
  return error;
}

template <class Val>
bool Future<Val>::is_emitter_error() const {
  Mutex::Lock lock(mutex);
  return emitter_error;
}

} // namespace Thread

#endif // FUTURE_TCC