/* Copyright (C) 2005, 2007 and 2009 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the c++-gtk-utils
   sub-directory); if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file or uses macros or
inline functions (of any length) should by reason only of that
instantiation or use be subject to the restrictions of use in the GNU
Lesser General Public License.  With that in mind, the words "and
macros, inline functions and instantiations of templates (of any
length)" shall be treated as substituted for the words "and small
macros and small inline functions (ten lines or less in length)" in
the fourth paragraph of section 5 of that licence.  This does not
affect any other reason why object code may be subject to the
restrictions in that licence (nor for the avoidance of doubt does it
affect the application of section 2 of that licence to modifications
of the source code in this file).

*/

#ifndef GOBJ_HANDLE_H
#define GOBJ_HANDLE_H

/**
 * @class GobjHandle gobj_handle.h c++-gtk-utils/gobj_handle.h
 * @brief This is a handle for managing the reference count of
 * GObjects.
 * @sa MainWidgetBase
 *
 * This is a class which manages the reference count of GObjects.  It
 * does not maintain its own reference count, but interfaces with that
 * kept by the glib object system.
 *
 * GobjHandles are most useful to manage GObjects which are not also
 * GtkObjects or GInitiallyUnowned objects - GtkObjects and
 * GInitiallyUnowned objects have initial floating references which
 * will result in them being automatically managed by the container in
 * which they are held.  Nonetheless, GobjHandles can be used to hold
 * GtkObjects and GInitiallyUnowned objects, as the constructor of a
 * GobjHandle which takes a pointer will automatically take ownership
 * of a newly created GtkObject or GInitiallyUnowned object, by
 * calling g_object_ref_sink().  Plain GObjects do not need to be sunk
 * to be owned by the GobjHandle.
 *
 * Note that g_object_ref_sink() is not called by the constructor
 * taking a pointer if the floating reference has already been sunk,
 * so if that constructor is passed an object already owned by a GTK+
 * container it will be necessary to call g_object_ref() on it
 * explicitly.  This behaviour will ensure that the handle behaves the
 * same whether it is holding a plain GObject, or it is holding a
 * GInitiallyUnowned/GtkObject object.  Generally however, where an
 * object is already owned by a container, the object should be passed
 * by another handle - ie by the copy constructor (or by the
 * assignment operator), as those always increment the reference count
 * automatically.
 *
 * In other words, invoke the constructor taking a pointer only with a
 * newly created object (whether a GObject, GInitiallyUnowned or
 * GtkObject object), and everything else will take care of itself. In
 * this respect, GobjHandles work the same way as conventional shared
 * pointer implementations managing objects allocated on free store.
 * The same applies to the reset() method.  Because any GTK+
 * containers themselves increment the reference count of a GObject or
 * GtkObject where they need to take ownership, an already-managed
 * object held by a GobjHandle can safely be passed by pointer to a
 * GTK+ container, and that is one of the intended usages.
 *
 * For that purpose, although the class has operator*() and
 * operator->() dereferencing operators, and so has normal smart
 * pointer functionality, as it is intended for use with the normal C
 * GObject/pango/GTK+ interfaces, ordinary use would involve passing
 * the handle to a function taking a pointer by means of the
 * operatorT*() type conversion operator (which returns the underlying
 * pointer), or by explicitly calling the get() method to obtain the
 * underlying pointer.
 *
 * The principle intended usage of GobjHandle is to automatically
 * handle GObject reference counts and therefore to make GObjects
 * exception-safe, but they also permit GObjects/GtkObjects to be kept
 * in standard C++ containers.
 *
 * As of glib-2.8, g_object_ref() and g_object_unref() are thread
 * safe, so with glib-2.8 or greater there can be different GobjHandle
 * instances in different threads referencing the same GObject object.
 * Of course, if that is done, this does not affect the need (or
 * otherwise) in the particular use in question to lock anything other
 * than the reference count - say when accessing the referenced object
 * itself in different threads.
 */

#include <glib-object.h>

template <class T> class GobjHandle {

  T* obj_p;

  void unreference() {
    if (obj_p) g_object_unref(G_OBJECT(obj_p));
  }

  void reference() {
    if (obj_p) g_object_ref(G_OBJECT(obj_p));
  }

public:
 /**
  * The constructor does not throw.  g_object_ref_sink() is called if
  * the managed object has a floating reference.
  * @param ptr The object which the GobjHandle is to manage (if any).
  * @note The object passed should not normally be owned by a GTK+
  * container or managed by any other GobjHandle object.  If it is,
  * g_object_ref() must be called explicitly by the user code.
  */
  explicit GobjHandle(T* ptr = 0) {
    obj_p = ptr;

    // if an object with a floating reference has been passed to this constructor,
    // take ownership of it
    if (obj_p && g_object_is_floating(G_OBJECT(obj_p))) {
      g_object_ref_sink(G_OBJECT(obj_p));
    }
  }

 /**
  * Causes the handle to cease to manage its managed object and
  * decrements the reference count, so destroying it if the reference
  * count thereby becomes 0.  If the argument passed is not NULL, the
  * handle will manage the new object passed and g_object_ref_sink()
  * is called if the new object has a floating reference.  This method
  * does not throw.
  * @param ptr NULL (the default), or a new object to manage.
  * @note The new object passed should not normally be owned by a GTK+
  * container or managed by any other GobjHandle object.  If it is,
  * g_object_ref() must be called explicitly by the user code.
  */
  void reset(T* ptr = 0) {
    
    unreference();
    obj_p = ptr;

    // if an object with a floating reference has been passed to this method,
    // take ownership of it
    if (obj_p && g_object_is_floating(G_OBJECT(obj_p))) {
      g_object_ref_sink(G_OBJECT(obj_p));
    }
  }

 /**
  * The copy constructor does not throw.  It increments the reference
  * count of the managed object.
@param gobj_ptr The handle to be
  * copied.
  */
  GobjHandle(const GobjHandle& gobj_ptr) {
    obj_p = gobj_ptr.obj_p;
    reference();
  }

 /**
  * This method does not throw.  It decrements the reference count of
  * the former managed object, so destroying it if the reference count
  * thereby becomes 0, and increments the reference count of the new
  * managed object.
  * @param
  * gobj_ptr The assignee.
  */
  GobjHandle& operator=(const GobjHandle& gobj_ptr) {

    // check whether we are already referencing this object -
    // if so make this a null op.  This will also deal with
    // self-assignment
    if (obj_p != gobj_ptr.obj_p) {

      // first unreference any object referenced by this handle
      unreference();

      // now inherit the GObject from the assigning handle
      // and reference it
      obj_p = gobj_ptr.obj_p;
      reference();
    }
    return *this;
  }

 /**
  * This method does not throw.
  * @return A pointer to the handled GObject.
  */
  T* get() const {return obj_p;}

 /**
  * This method does not throw.
  * @return A reference to the handled GObject.
  */
  T& operator*() const {return *obj_p;}

 /**
  * This method does not throw.
  * @return A pointer to the handled GObject.
  */
  T* operator->() const {return obj_p;}

 /**
  * This method does not throw.
  * @return A pointer to the handled GObject.
  */
  operator T*() const {return obj_p;}

 /**
  * The destructor does not throw.  It decrements the reference count
  * of the managed object, so destroying it if the reference count
  * thereby becomes 0.
  */
  ~GobjHandle() {unreference();}
};

#endif
