/* Copyright (C) 2005 to 2009 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the c++-gtk-utils
   sub-directory); if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file or uses macros or
inline functions (of any length) should by reason only of that
instantiation or use be subject to the restrictions of use in the GNU
Lesser General Public License.  With that in mind, the words "and
macros, inline functions and instantiations of templates (of any
length)" shall be treated as substituted for the words "and small
macros and small inline functions (ten lines or less in length)" in
the fourth paragraph of section 5 of that licence.  This does not
affect any other reason why object code may be subject to the
restrictions in that licence (nor for the avoidance of doubt does it
affect the application of section 2 of that licence to modifications
of the source code in this file).

*/

#ifndef MUTEX_H
#define MUTEX_H

#include <exception>
#include <pthread.h>
#include <time.h>

/*
  Define UTILS_COND_USE_MONOTONIC_CLOCK before the pre-processor
  parses this file if you don't want to use the default system clock
  for calls to Cond::timed_wait() - you will probably need to do this
  via the program configuration script by testing for the manifest
  constant CLOCK_MONOTONIC, that _POSIX_MONOTONIC_CLOCK is greater
  than 0 and that the pthread_condattr_setclock() function exists

  If UTILS_COND_USE_MONOTONIC_CLOCK is defined and a call to
  pthread_condattr_setclock() fails, the Cond object will remain
  uninitialised, and the exception CondSetClockError will be thrown by
  the constructor of the Cond object.
*/

/**
 * @file mutex.h
 * @brief Provides wrapper classes for pthread mutexes and condition
 * variables, and a scoped locking class for exception safe mutex
 * locking.
 * @note If UTILS_COND_USE_MONOTONIC_CLOCK is defined for the
 * pre-processor before this file is parsed, then a system monotonic
 * clock will be used in Thread::Cond::timed_wait() and
 * Thread::Cond::get_abs_time().
 */

namespace Thread {

struct CondError: public std::exception {
  virtual const char* what() const throw() {return "Thread::CondError";}
};

struct CondSetClockError: public std::exception {
  virtual const char* what() const throw() {return "Thread::CondSetClockError";}
};

struct MutexError: public std::exception {
  virtual const char* what() const throw() {return "Thread::MutexError";}
};

struct RecMutexError: public std::exception {
  virtual const char* what() const throw() {return "Thread::RecMutexError";}
};

class Cond;

/**
 * @class Mutex mutex.h c++-gtk-utils/mutex.h
 * @brief A wrapper class for pthread mutexes.
 * @sa Thread Mutex::Lock Cond RecMutex
 *
 * This class can be used interchangeably with threads started with
 * GThread and by this library, as both glib and this library use
 * pthreads underneath on POSIX and other unix-like OSes.  Unlike with
 * GMutex, there is no separate class for static mutexes.  That is not
 * necessary, as these Mutex objects can be constructed statically as
 * well as dynamically and there is no need to call g_thread_init()
 * before they are constructed.  (If created as a static object in
 * global scope, it will not be possible to catch Thread::MutexError
 * thrown by its constructor, but if a static global mutex throws
 * there is nothing that could be done anyway except abort, and it
 * would show that the pthreads installation is seriously defective.)
 */

class Mutex {
  pthread_mutex_t pthr_mutex;
  
  // mutexes cannot be copied
  Mutex(const Mutex&);
  Mutex& operator=(const Mutex&);
public:
  class Lock;
  friend class Cond;

/**
 * Locks the mutex and acquires ownership.  Blocks if already locked
 * until it becomes free.  It is not a cancellation point.  It does
 * not throw.  It is thread safe.
 * @return 0 if successful, otherwise the pthread mutex error number.
 */
  int lock() {return pthread_mutex_lock(&pthr_mutex);}

/**
 * Tries to lock the mutex and acquire ownership, but returns
 * immediately if it is already locked with value EBUSY.  It is not a
 * cancellation point.  It does not throw.  It is thread safe.
 * @return 0 if successful, otherwise EBUSY or other pthread mutex
 * error number.
 */
  int trylock() {return pthread_mutex_trylock(&pthr_mutex);}

/**
 * Unlocks a locked mutex owned by the calling thread and relinquishes
 * ownership.  It is not a cancellation point.  It does not throw.  It
 * is thread safe.

 * @return 0 if successful, otherwise the pthread mutex error number.  
 */
  int unlock() {return pthread_mutex_unlock(&pthr_mutex);}

/**
 * Initialises the pthread mutex.  It is not a cancellation point.
 * @exception Thread::MutexError Throws this exception if initialisation of
 * the mutex fails.  (It is often not worth checking for this, as it
 * means either memory is exhausted or pthread has run out of other
 * resources to create new mutexes.)
 */
  Mutex();

/**
 *  Destroys the pthread mutex.  It is not a cancellation point.  It
 *  does not throw.
 */
  ~Mutex() {pthread_mutex_destroy(&pthr_mutex);}
};

// used as a second argument to Lock::Lock() in cases where the
// mutex has already been locked (say by Mutex::trylock())
enum Locked {locked};

/**
 * @class Mutex::Lock mutex.h c++-gtk-utils/mutex.h
 * @brief A scoped locking class for exception safe Mutex locking.
 * @sa Mutex Thread Cond
 */

class Mutex::Lock {
  Mutex& mutex;

  // locks cannot be copied
  Lock(const Mutex::Lock&);
  Mutex::Lock& operator=(const Mutex::Lock&);
public:
  friend class Cond;

/**
 * Calls Mutex::lock(), and so locks the mutex and acquires ownership.
 * It blocks if already locked until the mutex becomes free.  This
 * method should normally only be called if a previous call has been
 * made to Mutex::Lock::unlock() (that is, where the thread owning the
 * Mutex::Lock object has temporarily allowed another thread to take
 * the mutex concerned).  It is not a cancellation point.  It does not
 * throw.  It is thread safe.
 * @return 0 if successful, otherwise the pthread mutex error number.
 */
  int lock() {return mutex.lock();}

/**
 * Calls Mutex::trylock(), and so tries to lock the mutex and acquire
 * ownership, but returns immediately if it is already locked with
 * value EBUSY.  This method should normally only be called if a
 * previous call has been made to Mutex::Lock::unlock() (that is,
 * where the thread owning the Mutex::Lock object has temporarily
 * allowed another thread to take the mutex concerned).  It is not a
 * cancellation point.  It does not throw.  It is thread safe.
 * @return 0 if successful, otherwise EBUSY or other pthread mutex
 * error number.
 */
  int trylock() {return mutex.trylock();}

/**
 * Calls Mutex::unlock(), and so unlocks a locked mutex owned by the
 * calling thread and relinquishes ownership (so temporarily allowing
 * another thread to take the mutex).  This method should normally
 * only be called if it is to be followed by a call to
 * Mutex::Lock::lock() or a successful call to Mutex::Lock::trylock()
 * before the Mutex::Lock object concerned goes out of scope
 * (otherwise Mutex::Lock's destructor will attempt to unlock an
 * already unlocked mutex or a mutex of which another thread has by
 * then taken ownership - Mutex::Lock objects do not maintain state).
 * It is not a cancellation point.  It does not throw.  It is thread
 * safe.
 * @return 0 if successful, otherwise the pthread mutex error number.  
 */
  int unlock() {return mutex.unlock();}

/**
 * This constructor locks the mutex passed to it.  It is not a
 * cancellation point.  It does not throw.
 * @param mutex_ The mutex to be locked.
 */
  Lock(Mutex& mutex_): mutex(mutex_) {mutex.lock();}

/**
 * This constructor takes an already locked mutex (say as a result of
 * Mutex::try_lock()), and takes ownership of it.  It is not a
 * cancellation point.  It does not throw.
 * @param mutex_ The mutex to be managed by this object.
 * @param tag Pass the Thread::locked enum tag to this parameter.
 */
  Lock(Mutex& mutex_, Locked tag): mutex(mutex_) {}

/**
 * The destructor unlocks the owned mutex.  It is not a cancellation
 * point.  It does not throw.
 */
  ~Lock() {mutex.unlock();}
};

/**
 * @class Cond mutex.h c++-gtk-utils/mutex.h
 * @brief A wrapper class for pthread condition variables.
 * @sa Thread Mutex Mutex::Lock
 */

class Cond {
  pthread_cond_t cond;

  // Cond cannot be copied
  Cond(const Cond&);
  Cond& operator=(const Cond&);
public:

/**
 * Unblock at least one thread waiting on this condition variable.
 * Can be called by any thread. It is not a cancellation point.  Does
 * not throw.
 * @return 0 if successful, otherwise the pthread error number.  
 */
  int signal() {return pthread_cond_signal(&cond);}

/**
 * Unblocks all threads waiting on this condition variable, which
 * acquire the mutex in an order determined by the scheduling policy.
 * Can be called by any thread. It is not a cancellation point.  Does
 * not throw.
 * @return 0 if successful, otherwise the pthread error number.  
 */
  int broadcast() {return pthread_cond_broadcast(&cond);}

/**
 * Waits on this condition variable until unblocked or otherwise
 * awoken.  It must be called by the thread which owns the mutex.
 * Re-acquires the mutex when awoken.  It is a cancellation point.
 * This method is cancellation safe even if the stack does not unwind
 * on cancellation (but if the stack does not unwind some other way of
 * destroying this object on cancellation is required, such as by
 * having it allocated on freestore and deleted in a cancellation
 * clean-up handler).  This method does not throw.
 * @param mutex The locked mutex associated with the wait which is
 * re-acquired on being awoken.
 * @return 0 if successful, otherwise the pthread error number.
 * @note pthread condition variables can, as a matter of design, awake
 * spontaneously (and Cond::signal() may awaken more than one thread).
 * Therefore the relevant condition should be tested in a while loop
 * and not in an if block.
 */
  int wait(Mutex& mutex) {return pthread_cond_wait(&cond, &mutex.pthr_mutex);}

/**
 *  Does the same as Cond::wait(Mutex&), except that as a convenience
 *  it will take a Mutex::Lock object handling the Mutex object as an
 *  alternative to passing the Mutex object itself.
 */
  int wait(Mutex::Lock& lock) {return wait(lock.mutex);}

/**
 * Waits on this condition variable until unblocked or otherwise
 * awoken (in which case it re-acquires the mutex), or until the
 * timeout expires (in which case it returns with ETIMEDOUT).  It must
 * be called by the thread which owns the mutex. Re-acquires the mutex
 * when awoken or timing out.  It is a cancellation point.  This
 * method is cancellation safe even if the stack does not unwind on
 * cancellation (but if the stack does not unwind some other way of
 * destroying this object on cancellation is required, such as by
 * having it allocated on freestore and deleted in a cancellation
 * clean-up handler).  This method does not throw.
 * @param mutex The locked mutex associated with the wait which is
 * re-acquired on being awoken or timing out.
 * @param abs_time The time at which the wait will unblock if not
 * previously awoken.  A suitable value can be obtained by calling
 * the get_abs_time() function.
 * @return 0 if successful, otherwise ETIMEDOUT or other pthread error
 * number.
 * @note 1. pthread condition variables can, as a matter of design,
 * awake spontaneously (and Cond::signal() may awaken more than one
 * thread).  Therefore the relevant condition should be tested in a
 * while loop and not in an if block.
 * @note 2. If UTILS_COND_USE_MONOTONIC_CLOCK is defined for the
 * pre-processor before this file is parsed, then this condition
 * variable will use a system monotonic clock in this method
 * and in Cond::get_abs_time().
 */
  int timed_wait(Mutex& mutex, const timespec& abs_time) {
    return pthread_cond_timedwait(&cond, &mutex.pthr_mutex, &abs_time);
  }

/**
 *  Does the same as Cond::timed_wait(Mutex&, const timespec&), except
 *  that as a convenience it will take a Mutex::Lock object handling
 *  the Mutex object as an alternative to passing the Mutex object
 *  itself.
 */
  int timed_wait(Mutex::Lock& lock,
		 const timespec& abs_time) {return timed_wait(lock.mutex, abs_time);}

/**
 * This is a utility function that inserts into a timespec structure
 * the current time plus a given number of milliseconds ahead, which
 * can be applied to a call to Cond::timed_wait().  It does not throw.
 * It is thread-safe.
 * @param ts A timespec object into which the result of current time +
 * millisec will be placed.
 * @param millisec The number of milliseconds ahead of current time to
 * which the timespec object will be set.
 * @note If UTILS_COND_USE_MONOTONIC_CLOCK is defined for the
 * pre-processor before this file is parsed, then condition variables
 * will use a system monotonic clock in this method and
 * Cond::timed_wait().
 */
  static void get_abs_time(timespec& ts, unsigned int millisec) {
#ifdef UTILS_COND_USE_MONOTONIC_CLOCK
    clock_gettime(CLOCK_MONOTONIC, &ts);
#else
    clock_gettime(CLOCK_REALTIME, &ts);
#endif
    unsigned long nanosec = ts.tv_nsec + ((millisec % 1000) * 1000000);
    ts.tv_sec += millisec/1000 + nanosec/1000000000;
    ts.tv_nsec = nanosec % 1000000000;
  }

/**
 * Initialises the pthread condition variable.  It is not a
 * cancellation point.
 * @exception Thread::CondSetClockError Throws this exception if the
 * system does not support monotonic clocks and
 * UTILS_COND_USE_MONOTONIC_CLOCK is defined.  (If this has been
 * checked beforehand there should be no need to check for this
 * exception.)
 * @exception Thread::CondError Throws this exception if
 * initialisation of the condition variable fails, except because the
 * system does not support monotonic clocks when
 * UTILS_COND_USE_MONOTONIC_CLOCK is defined.  (It is often not worth
 * checking for CondError, as it means either memory is exhausted or
 * pthread has run out of other resources to create new condition
 * variables.)
 * @note If UTILS_COND_USE_MONOTONIC_CLOCK is defined for the
 * pre-processor before this file is parsed (it is an inline
 * function), then this condition variable will use a system monotonic
 * clock in Cond::timed_wait() and Cond::get_abs_time().
 */
  Cond() {
#ifdef UTILS_COND_USE_MONOTONIC_CLOCK
    pthread_condattr_t attr;
    if (pthread_condattr_init(&attr))
      throw CondError();
    if (pthread_condattr_setclock(&attr, CLOCK_MONOTONIC)) {
      pthread_condattr_destroy(&attr);
      throw CondSetClockError();
    }
    if (pthread_cond_init(&cond, &attr)) {
      pthread_condattr_destroy(&attr);
      throw CondError();
    }
    pthread_condattr_destroy(&attr);
#else
    if (pthread_cond_init(&cond, 0))
      throw CondError();
#endif
  }

/**
 * Destroys the pthread condition variable. It is not a cancellation
 * point.  The destructor does not throw.
 */
  ~Cond(void) {pthread_cond_destroy(&cond);}
};

/**
 * @class RecMutex mutex.h c++-gtk-utils/mutex.h
 * @brief A wrapper class for pthread mutexes which provides a
 * recursive mutex.
 * @sa Thread RecMutex::Lock Mutex
 *
 * This class can be used interchangeably with threads started with
 * GThread and by this library, as both glib and this library use
 * pthreads underneath on POSIX and other unix-like OSes.  Unlike with
 * GRecMutex, there is no separate class for static mutexes.  That is
 * not necessary, as these RecMutex objects can be constructed
 * statically as well as dynamically and there is no need to call
 * g_thread_init() before they are constructed.  (If created as a
 * static object in global scope, it will not be possible to catch
 * Thread::MutexError or Thread::RecMutexError thrown by its
 * constructor, but if a static global mutex throws there is nothing
 * that could be done anyway except abort, and it would show that the
 * pthreads installation is seriously defective.)
 */

class RecMutex {
  pthread_mutex_t pthr_mutex;
  
  // mutexes cannot be copied
  RecMutex(const RecMutex&);
  RecMutex& operator=(const RecMutex&);
public:
  class Lock;

/**
 * Locks the mutex and acquires ownership.  Blocks if already locked
 * until it becomes free, unless the calling thread already holds the
 * lock, in which case it increments the lock count and returns
 * immediately.  It is not a cancellation point.  It does not throw.
 * It is thread safe.
 * @return 0 if successful, otherwise the pthread mutex error number.
 */
  int lock() {return pthread_mutex_lock(&pthr_mutex);}

/**
 * Tries to lock the mutex and acquire ownership, but returns
 * immediately if it is already locked with value EBUSY unless the
 * calling thread already holds the lock, in which case it returns
 * normally and increments the lock count.  It is not a cancellation
 * point.  It does not throw.  It is thread safe.
 * @return 0 if successful, otherwise EBUSY or other pthread mutex
 * error number.
 */
  int trylock() {return pthread_mutex_trylock(&pthr_mutex);}

/**
 * Unlocks a locked mutex owned by the calling thread and either
 * relinquishes ownership (if the mutex has not been recursively
 * locked) or decrements the lock count (if it has).  It is not a
 * cancellation point.  It does not throw.  It is thread safe.
 * @return 0 if successful, otherwise the pthread mutex error number.  
 */
  int unlock() {return pthread_mutex_unlock(&pthr_mutex);}

/**
 * Initialises the pthread mutex.  It is not a cancellation point.
 * @exception Thread::RecMutexError Throws this exception if the
 * system does not support recursive mutexes.  (If this has been
 * checked beforehand there should be no need to check for this
 * exception.)
 * @exception Thread::MutexError Throws this exception if
 * initialisation of the mutex fails, except because the system does
 * not support recursive mutexes.  (It is often not worth checking for
 * MutexError, as it means either memory is exhausted or pthread has
 * run out of other resources to create new mutexes.)
 */
  RecMutex();

/**
 *  Destroys the pthread mutex.  It is not a cancellation point.  It
 *  does not throw.
 */
  ~RecMutex() {pthread_mutex_destroy(&pthr_mutex);}
};

/**
 * @class RecMutex::Lock mutex.h c++-gtk-utils/mutex.h
 * @brief A scoped locking class for exception safe RecMutex locking.
 * @sa RecMutex Thread
 */

class RecMutex::Lock {
  RecMutex& mutex;

  // locks cannot be copied
  Lock(const RecMutex::Lock&);
  RecMutex::Lock& operator=(const RecMutex::Lock&);
public:

/**
 * This calls RecMutex::lock(), and so locks the mutex and acquires
 * ownership.  It blocks if already locked until the mutex becomes
 * free, unless the calling thread already holds the lock, in which
 * case it increments the lock count and returns immediately.  This
 * method should normally only be called if a previous call has been
 * made to RecMutex::Lock::unlock() (that is, where the thread owning
 * the RecMutex::Lock object has temporarily allowed another thread to
 * take the mutex concerned if it is not recursively locked).  It is
 * not a cancellation point.  It does not throw.  It is thread safe.
 * @return 0 if successful, otherwise the pthread mutex error number.
 */
  int lock() {return mutex.lock();}

/**
 * This calls RecMutex::trylock(), and so tries to lock the mutex and
 * acquire ownership, but returns immediately if it is already locked
 * with value EBUSY unless the calling thread already holds the lock,
 * in which case it returns normally and increments the lock count.
 * This method should normally only be called if a previous call has
 * been made to RecMutex::Lock::unlock() (that is, where the thread
 * owning the RecMutex::Lock object has temporarily allowed another
 * thread to take the mutex concerned if it is not recursively
 * locked).  It is not a cancellation point.  It does not throw.  It
 * is thread safe.
 * @return 0 if successful, otherwise EBUSY or other pthread mutex
 * error number.
 */
 int trylock() {return mutex.trylock();}

/**
 * This calls RecMutex::unlock(), and so unlocks a locked mutex owned
 * by the calling thread and either relinquishes ownership (if the
 * mutex has not been recursively locked), so temporarily allowing
 * another thread to take the mutex, or decrements the lock count (if
 * it has been recursively locked).  This method should normally only
 * be called if it is to be followed by a call to
 * RecMutex::Lock::lock() or a successful call to
 * RecMutex::Lock::trylock() before the RecMutex::Lock object
 * concerned goes out of scope (otherwise RecMutex::Lock's destructor
 * will attempt to decrement the lock count of a mutex which already
 * has a lock count of 0 or which another thread has by then taken
 * ownership or leave the lock count in an unbalanced condition -
 * RecMutex::Lock objects do not maintain state).  It is not a
 * cancellation point.  It does not throw.  It is thread safe.
 * @return 0 if successful, otherwise the pthread mutex error number.  
 */
 int unlock() {return mutex.unlock();}

/**
 * This constructor locks the mutex passed to it.  See
 * RecMutex::lock() for a description of the outcomes.  It is not a
 * cancellation point.  It does not throw.
 * @param mutex_ The mutex to be locked.
 */
  Lock(RecMutex& mutex_): mutex(mutex_) {mutex.lock();}

/**
 * This constructor takes an already locked mutex (say as a result of
 * RecMutex::try_lock()), and takes ownership of it.  It is not a
 * cancellation point.  It does not throw.
 * @param mutex_ The mutex to be managed by this object.
 * @param tag Pass the Thread::locked enum tag to this parameter.
 */
  Lock(RecMutex& mutex_, Locked tag): mutex(mutex_) {}

/**
 * The destructor unlocks the owned mutex.  See RecMutex::unlock() for
 * a description of the outcomes.  It is not a cancellation point.  It
 * does not throw.
 */
  ~Lock() {mutex.unlock();}
};

} // namespace Thread

#endif
